# simall ############################################################
simall <-
function( mods, nd, N, 
          int=mean(diff(nd[,1])) ) # Assuming time variable is first in nd
{
# from a list of models for cause-specific rates and a common
# prediction frame generate N samples of predicted rates
# it is assumed that the nd refer to midpoints of intervals 
res <- NArray( list( time = rownames(nd),
                      mod = names(mods),
                      sim = 0:N ) )
for( i in 1:length(mods) )
   { 
   res[,i,1    ] <- ci.lin( mods[[i]], nd )[,1]
   res[,i,1:N+1] <- ci.lin( mods[[i]], nd, sample=N )
   }
res <- exp(res)
attr( res, "int" ) <- int
res
}

# mkprob ############################################################
mkprob <-
function( rtab, int )
{
# rtab is assumed to be a matrix with rates in (named) columns at int
# equidistance assumed computed at times int/2, 3*int/2 etc. so rtab
# is assumed T x C  

# integrated intensites at interval boundaries, so (T+1) x C
Rtab <- rbind(0,apply( rtab, 2, cumsum ))*int
    
# survival function (also the first column in Pr)
Pr <- Sv <- exp( -apply( Rtab, 1, sum ) )

# convenience function 
mp <- function(x) x[-1]-diff(x)/2    
    
# cumulative risks note we are taking the Sv at midpoint of interval
for( i in 1:ncol(rtab) ) 
     Pr <- cbind( Pr, c(0,cumsum( rtab[,i]*mp(Sv) )*int) ) 
colnames( Pr ) <- c("Surv", colnames(rtab) )
# Pr is now a (T+1) x (C+1) matrix
Pr 
}

# mksm ############################################################
mksm <-
function( mods, nd, int=mean(diff(nd[,1])), N )
{
res <- simall( mods, nd, N )
# res is now C x T x (N+1)

dn <- dimnames( mkprob( res[,,1], int ) )
prob <- NArray( c(dn,list(c(0:N))) )
for( i in 0:N+1 ) prob[,,i] <- mkprob( res[,,i], int )
dimnames(prob)[[1]] <- 0:nrow(nd) * int
# prob is now (C+1) x (T+1) x (N+1)
# C+1 because we added surv by using mkprob, 
# T+1 because we now have classification by boundaries and not
#     midpoints (mkprob)
# N+1 because the first entry is the fitted values from the model 

# Finally we carry the interval length over
attr( prob, 'int' ) <- int
prob
}

# mnqt ############################################################
mnqt <- 
function( x, alpha=0.05 ) 
# extracts the statistics based on the model parameters (1st
# simulation) and the mean and the relevant quantiles across the
# simulated samples
       c( model=     x[ 1] , 
           mean=mean(x[-1]),
            quantile(x[-1],
                     c(0.5,alpha/2,1-alpha/2),
                     na.rm=TRUE) )

# prci ############################################################
prci <-
function( prob, 
         alpha = 0.05, 
          clev = 1-alpha )
{
# statistics for the state probabilities 
pr <- apply( prob, 1:2, mnqt, alpha )
}

# prst ############################################################
prst <-
function( prob, 
          perm = 1:ncol(prob), 
         alpha = 0.05, 
          clev = 1-alpha ) 
{
# statistiscs for the stacked probabilities    
# cumulate over permuted states and get the dimensions in the right order
prpm <- aperm( apply( prob[,perm,], c(1,3), cumsum ), c(2,1,3) )
# then derive the estimates and the confidence intervals for the
# cumulative probabilities
apply( prpm, 1:2, mnqt, alpha )        
}

# prsj ############################################################
prsj <-
function( prob, 
         alpha = 0.05, 
          clev = 1-alpha,
           int = attr(prob,'int') )
{
# Statistics for the sojourn times in (0,t)
# To get the sojourn times in each state as a function of time
# Sojourn times in each of the absorbing states is what we could call
# the years lost to that cause.
sj <- apply( prob, 2:3, cumsum ) * int
apply( sj, 1:2, mnqt, alpha )
}

# ci.cmpr ##########################################################
ci.prob <-
ci.cmpr <-
function( mods, nd, N, 
          int = mean(diff(nd[,1])),
         what = "prob", # stack (st),
                        # sojourn (so,sj),
                        # sample (sa,sm)
         perm = 1:(length(mods)+1),
        alpha = 0.05 ) 
{
# Generate the simulated probabilities (T+1)x(C+1)x(N+1)
smpr <- mksm( mods = mods,
                nd = nd,
                 N = N,
               int = int )
# Confidence intervals for the selected statistics
wh <- tolower(substr(what,1,2))
rs <- list( probs  =if("pr" %in% wh) prci(smpr,alpha=alpha),
            stack  =if("st" %in% wh) prst(smpr,alpha=alpha,perm=perm),
            sojourn=if("so" %in% wh |
                       "sj" %in% wh) prsj(smpr,alpha=alpha),
            sample =if("sa" %in% wh |
                       "sm" %in% wh)      smpr )
# no NULL entries in list
rs <- rs[which(!sapply(rs,is.null))]
# return object if list has length 1
if( length(rs)==1 ) rs <- rs[[1]]
rs
}    
